# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2012 m@n <rmaneschi@gmail.com>
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE
from __future__ import print_function

import optparse

import gettext
from gettext import gettext as _
gettext.textdomain('showmyfaves')

import asyncore
import collections
import logging
import socket
import sys
import os
import json
import threading
import gobject

import ShowmyfavesModel
import ShowmyfavesUtil

class Host(asyncore.dispatcher):

    log = logging.getLogger('Host')

    def __init__(self):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.config = ShowmyfavesUtil.ConfigLoader.load()
        self.log.info("Server started on "+str(self.config.HOST_URL)+":"+str(self.config.HOST_PORT))
        self.bind((self.config.HOST_URL, self.config.HOST_PORT))
        self.listen(1)
        self.remote_clients = []

    def handle_accept(self):
        socket, addr = self.accept() # For the remote client.
        self.log.info('Accepted client at %s', addr)
        self.remote_clients.append(RemoteClient(self, socket, addr))

    def broadcast(self, message):
        self.log.info('Broadcasting message: %s', message)
        for remote_client in self.remote_clients:
            remote_client.sendCommand(message)
            
class RemoteClient(asyncore.dispatcher):

    """Wraps a remote client socket."""

    log = logging.getLogger('RemoteClient')

    def __init__(self, host, socket, address):
        asyncore.dispatcher.__init__(self, socket)
        self.host = host
        self.outbox = collections.deque()
        self.address = address
        self.commands = ServerCommands(self)

    def sendToClient(self, message):
        self.outbox.append(message.encode('utf-8'))

    def handle_read(self):
        client_message = self.socket.recv(self.host.config.MAX_MESSAGE_LENGTH)
        if client_message == "":
            self.log.info('Disconnect %s', self.address)
            self.close()
            self.host.remote_clients.remove(self)
        else:
            self.commands.parse(client_message)
#            self.host.broadcast(client_message)

    def writable(self):
        return (len(self.outbox) > 0)

    def handle_write(self):
        message = self.outbox.popleft()
        if not message is None:
#            if len(message) > self.host.config.MAX_MESSAGE_LENGTH:
#                raise ValueError('Message too long')
            self.send(message)
        
    def handle_close(self):
        self.close()
            
class ServerCommands():
    
    log = logging.getLogger('ServerCommands')
    
    def __init__(self, listener):
        pathname = os.path.dirname(__file__)
        sys.path.append(pathname + "/../parsers")
        self.parsers = os.listdir(pathname + "/../parsers")
        sys.path.append(pathname + "/../sorts")
        self.sorts = os.listdir(pathname + "/../sorts")
        self.listener = listener
        
    def sendToClient(self, command, commandId, data):
        res = {'cmd':command, 'id':commandId, 'data':data}
        message = json.dumps(res, cls=MyEncoder)
        # prevent cumulData on network, so add eof
        message += "#||;"
        self.listener.sendToClient(message)
        
    def parse(self, message):
        self.log.info('Command: %s', message)
        data = json.loads(message)
        command = data["cmd"]
        params = {}
        if "params" in data:
            params = data["params"]
        if command == "close":
            return
        elif command == "getBookmarks":
            res = self.getBookmarks(params)
            #send data to client
            for r in res:
                self.sendToClient("getBookmarks", data["id"], r)
        elif command == "getNbBookmarks":
            self.getNbBookmarks(data["id"], params)
        elif command == "addBookmark":
            pass
        elif command == "getTags":
            self.getTags(data["id"])
        elif command == "getParsers":
            self.getParsers(data["id"])
        elif command == "getSorts":
            self.getSorts(data["id"])
        else:
            self.listener.sendCommand('{"error":"command not found, command available : getBookmarks"}') 
        
    def getBookmarks(self, params={}):
        #parsers
        res = self.getParsersBookmarks(params)
        #sorts
        res = self.sortBookmarks(res, params)
        return res
            
    def getNbBookmarks(self, commandId, params={}):
        #parsers
        res = self.getParsersBookmarks(params)
        #send data to client
        self.sendToClient("getNbBookmarks", commandId, len(res))

    def getTags(self, commandId):
        res = []
        bookmarks = self.getBookmarks({"parser":[]})
        for bookmark in bookmarks:
            for tag in bookmark.tags:
                t = ShowmyfavesModel.Tag.getTag(tag.type, tag.name)
                if not t in res:
                    res.append(t)
        for r in res:
            self.sendToClient("getTags", commandId, r)
        
    def getParsers(self, commandId):
        res = []
        for fname in self.parsers:
            if fname.endswith("py") and not fname == "ParsersUtil.py":
                res.append(fname.rstrip(".py"))
        self.sendToClient("getParsers", commandId, res)
        
    def getSorts(self, commandId):
        res = []
        for fname in self.sorts:
            if fname.endswith("py"):
                res.append(fname.rstrip(".py"))
        self.sendToClient("getSorts", commandId, res)
        
    def getParsersBookmarks(self, params):
        res = []
        for fname in self.parsers:
            if fname.endswith("py") and not fname == "__init__.py" and not fname == "ParsersUtil.py":
                parserName = fname.rstrip(".py")
                if params and "parser" in params and (parserName in params["parser"] or len(params["parser"]) == 0):
                    module = __import__(parserName)
                    filters = None
                    res = res + module.Parse(params)
        return res
        
    def sortBookmarks(self, res, params):
        if params and "sort" in params:
            for fname in self.sorts:
                if fname.endswith("py"):
                    sortName = fname.rstrip(".py")
                    if sortName in params["sort"]:
                        module = __import__(sortName)
                        res = module.Sort(res, params["reverseSort"])
        return res
        
class MyEncoder(json.JSONEncoder):
    def default(self, obj):
        return obj.__dict__

class FakeServer(threading.Thread):

    log = logging.getLogger("FakeServer")

    def __init__(self, listener=None):
        threading.Thread.__init__(self)
        self.listner = listener
        self.serverCommand = ServerCommands(self)
        
    def sendCommand(self, command):
        self.serverCommand.parse(command)
        
    def sendToClient(self, command):
        data = json.loads(command)
        if self.listner != None:
            if "error" in data:
                gobject.idle_add(self.listner.error, data["error"])
            else:
                if data["cmd"] == "getBookmarks":
                    bookmarks = []
                    for bookmark in data["data"]:
                        d = ShowmyfavesModel.Bookmark().fromDict(**bookmark)
                        bookmarks.append(d)
                    gobject.idle_add(self.listner.setBookmarks, bookmarks)
                elif data["cmd"] == "getParsers":
                    gobject.idle_add(self.listner.setParsers, data["data"])
                elif data["cmd"] == "getSorts":
                    gobject.idle_add(self.listner.setSorts, data["data"])
                elif message == "":
                    self.stop()
        else:
            self.log.info("No listner on client !!")
            
    def stop(self):
        self._Thread__stop()

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    host = Host()
    asyncore.loop()
    
